/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.options.cachemodelocal;

import org.jboss.cache.CacheFactory;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.UnitTestCacheFactory;
import org.jboss.cache.commands.ReplicableCommand;
import org.jboss.cache.commands.write.PutDataMapCommand;
import org.jboss.cache.commands.write.PutKeyValueCommand;
import org.jboss.cache.commands.write.RemoveKeyCommand;
import org.jboss.cache.commands.write.RemoveNodeCommand;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.util.TestingUtil;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.transaction.TransactionManager;
import java.util.HashMap;
import java.util.Map;

/**
 * Tests the cache mode local override in various scenarios.  To be subclassed to test REPL_SYNC, REPL_ASYNC, INVALIDATION_SYNC, INVALIDATION_ASYNC for Opt and Pess locking.
 * <p/>
 * Option.setCacheModeLocal() only applies to put() and remove() methods.
 *
 * @author <a href="mailto:manik AT jboss DOT org">Manik Surtani (manik AT jboss DOT org)</a>
 */
@Test(groups = {"functional"}, testName = "cachemodelocal.CacheModeLocalTestBase")
public abstract class CacheModeLocalTestBase
{
   // to be subclassed.
   protected Configuration.CacheMode cacheMode;
   protected String nodeLockingScheme;
   /**
    * set this to true if the implementing class plans to use an invalidating cache mode *
    */
   protected boolean isInvalidation;

   private ThreadLocal<CacheSPI<String, String>> cache1TL = new ThreadLocal<CacheSPI<String, String>>();
   private ThreadLocal<CacheSPI<String, String>> cache2TL = new ThreadLocal<CacheSPI<String, String>>();
   private ThreadLocal<NodeSPI<String, String>> root1TL = new ThreadLocal<NodeSPI<String, String>>();
   private ThreadLocal<NodeSPI<String, String>> root2TL = new ThreadLocal<NodeSPI<String, String>>();

   private final Fqn fqn = Fqn.fromString("/a");
   private final String key = "key";

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {

      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      // force a tear down if the test runner didn't run one before (happens in IDEA)
      if (cache1 != null || cache2 != null)
         tearDown();

      CacheFactory<String, String> instance = new UnitTestCacheFactory<String, String>();

      Configuration c = new Configuration();
      c.setClusterName("test");
      c.setStateRetrievalTimeout(1000);
      c.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      c.setNodeLockingScheme(nodeLockingScheme);
      c.setCacheMode(cacheMode);

      cache1 = (CacheSPI<String, String>) instance.createCache(c, false);
      cache1.start();

      c = new Configuration();
      c.setClusterName("test");
      c.setStateRetrievalTimeout(1000);
      c.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      c.setNodeLockingScheme(nodeLockingScheme);
      c.setCacheMode(cacheMode);

      cache2 = (CacheSPI<String, String>) instance.createCache(c, false);
      cache2.start();

      cache1TL.set(cache1);
      cache2TL.set(cache2);

      root1TL.set(cache1.getRoot());
      root2TL.set(cache2.getRoot());
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown()
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      TestingUtil.killCaches(cache1, cache2);
      cache1TL.set(null);
      cache2TL.set(null);
   }

   public void testPutKeyValue() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, key, "value");
      Thread.sleep(500);

      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));

      // cache 2 should not
      assertNull("Should be null", cache2.get(fqn, key));

      // now try again with passing the default options
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.put(fqn, key, "value");
      verifyReplication();
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));

      // cache 2 should as well
      if (!isInvalidation)
      {
         assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, key, "value2");
      Thread.sleep(500);
      assertEquals("value2", cache2.get(fqn, key));
      assertEquals("value", cache1.get(fqn, key));

      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      cache2.getInvocationContext().getOptionOverrides().reset();
      cache2.put(fqn, key, "value2");
      verifyReplication();
      assertEquals("value2", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         assertEquals("value2", cache1.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache1.get(fqn, key));
      }
   }

   public void testPutKeyValueViaNodeAPI() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();
      NodeSPI<String, String> root1 = root1TL.get();
      NodeSPI<String, String> root2 = root2TL.get();

      Node node1 = root1.addChild(fqn);
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node1.put(key, "value");
      Thread.sleep(500);
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));

      // cache 2 should not
      assertNull("Should be null", cache2.get(fqn, key));

      // now try again with passing the default options
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      cache1.getInvocationContext().getOptionOverrides().reset();
      node1.put(key, "value");
      verifyReplication();
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));

      // cache 2 should as well
      if (!isInvalidation)
      {
         assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      Node node2 = root2.addChild(fqn);
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node2.put(key, "value2");
      Thread.sleep(500);

      assertEquals("value2", cache2.get(fqn, key));
      assertEquals("value", cache1.get(fqn, key));

      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      cache2.getInvocationContext().getOptionOverrides().reset();
      node2.put(key, "value2");
      verifyReplication();
      assertEquals("value2", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         assertEquals("value2", cache1.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache1.get(fqn, key));
      }
   }

   public void testPutData() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      Map<String, String> map = new HashMap<String, String>();
      map.put(key, "value");

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, map);
      Thread.sleep(500);
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));
      // cache 2 should not
      assertNull("Should be null", cache2.get(fqn, key));

      // now try again with passing the default options
      registerReplicationCommand(cache2, PutDataMapCommand.class);
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      cache1.put(fqn, map);
      verifyReplication();
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));
      // cache 2 should as well
      if (!isInvalidation)
      {
         assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      map.put(key, "value2");
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, map);
      Thread.sleep(500);

      assertEquals("value2", cache2.get(fqn, key));
      assertEquals("value", cache1.get(fqn, key));

      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      cache2.getInvocationContext().getOptionOverrides().reset();
      cache2.put(fqn, key, "value2");
      verifyReplication();
      assertEquals("value2", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         assertEquals("value2", cache1.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache1.get(fqn, key));
      }
   }

   public void testPutDataViaNodeAPI() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();
      NodeSPI<String, String> root1 = root1TL.get();
      NodeSPI<String, String> root2 = root2TL.get();

      Map<String, String> map = new HashMap<String, String>();
      map.put(key, "value");

      Node node1 = root1.addChild(fqn);
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node1.putAll(map);
      Thread.sleep(500);
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));
      // cache 2 should not
      assertNull("Should be null", cache2.get(fqn, key));

      // now try again with passing the default options
      registerReplicationCommand(cache2, PutDataMapCommand.class);
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      node1.putAll(map);
      verifyReplication();
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));
      // cache 2 should as well
      if (!isInvalidation)
      {
         assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      Node node2 = root2.addChild(fqn);
      map.put(key, "value2");
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node2.putAll(map);
      Thread.sleep(500);

      assertEquals("value2", cache2.get(fqn, key));
      assertEquals("value", cache1.get(fqn, key));

      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      cache2.getInvocationContext().getOptionOverrides().reset();
      node2.put(key, "value2");
      verifyReplication();
      assertEquals("value2", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         assertEquals("value2", cache1.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache1.get(fqn, key));
      }
   }

   public void testRemoveNode() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      // put some stuff in the cache first
      // make sure we cleanup thread local vars.
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      cache1.getInvocationContext().setOptionOverrides(null);
      cache1.put(fqn, key, "value");
      verifyReplication();
      assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.removeNode(fqn);
      Thread.sleep(500);

      // should be removed in cache1
      assertNull("should be null", cache1.get(fqn, key));
      // Not in cache2
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      // replace cache entries
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      cache1.put(fqn, key, "value");
      verifyReplication();
      assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      // now try again with passing the default options
      registerReplicationCommand(cache2, RemoveNodeCommand.class);
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      cache1.removeNode(fqn);
      verifyReplication();

      // both should be null
      assertNull("should be null", cache1.get(fqn, key));
      assertNull("should be null", cache2.get(fqn, key));
   }

   public void testRemoveNodeViaNodeAPI() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();
      NodeSPI<String, String> root1 = root1TL.get();
      NodeSPI<String, String> root2 = root2TL.get();

      // put some stuff in the cache first
      // make sure we cleanup thread local vars.
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      cache1.getInvocationContext().setOptionOverrides(null);
      cache1.put(fqn, key, "value");
      assertEquals("value", cache1.get(fqn, key));
      verifyReplication();
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      root1.removeChild(fqn);
      Thread.sleep(500);

      // should be removed in cache1
      assertNull("should be null", cache1.get(fqn, key));
      // Not in cache2
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      // replace cache entries
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      cache1.put(fqn, key, "value");
      verifyReplication();
      assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      registerReplicationCommand(cache2, RemoveNodeCommand.class);
      // now try again with passing the default options
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      root1.removeChild(fqn);
      verifyReplication();

      // both should be null
      assertNull("should be null", cache1.get(fqn, key));
      assertNull("should be null", cache2.get(fqn, key));
   }

   public void testRemoveKey() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      // put some stuff in the cache first
      cache1.getInvocationContext().setOptionOverrides(null);
      cache1.put(fqn, key, "value");
      verifyReplication();
      assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.remove(fqn, key);
      Thread.sleep(500);

      // should be removed in cache1
      assertNull("should be null", cache1.get(fqn, key));
      // Not in cache2
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      // replace cache entries
      cache1.put(fqn, key, "value");
      verifyReplication();
      assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      registerReplicationCommand(cache2, RemoveKeyCommand.class);
      // now try again with passing the default options
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.remove(fqn, key);
      verifyReplication();

      // both should be null
      assertNull("should be null", cache1.get(fqn, key));
      assertNull("should be null", cache2.get(fqn, key));
   }

   public void testRemoveKeyViaNodeAPI() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();
      NodeSPI<String, String> root1 = root1TL.get();
      NodeSPI<String, String> root2 = root2TL.get();

      // put some stuff in the cache first
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      Node node1 = root1.addChild(fqn);
      cache1.getInvocationContext().setOptionOverrides(null);
      node1.put(key, "value");
      verifyReplication();
      assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node1.remove(key);
      Thread.sleep(500);

      // should be removed in cache1
      assertNull("should be null", cache1.get(fqn, key));
      // Not in cache2
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      // replace cache entries
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      node1.put(key, "value");
      verifyReplication();
      assertEquals("value", cache1.get(fqn, key));
      if (isInvalidation)
      {
         assertNull("Should be null", cache2.get(fqn, key));
      }
      else
      {
         assertEquals("value", cache2.get(fqn, key));
      }

      // now try again with passing the default options
      registerReplicationCommand(cache2, RemoveKeyCommand.class);
      cache1.getInvocationContext().getOptionOverrides().reset();
      node1.remove(key);
      verifyReplication();

      // both should be null
      assertNull("should be null", cache1.get(fqn, key));
      assertNull("should be null", cache2.get(fqn, key));
   }

   public void testTransactionalBehaviourCommit() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      TransactionManager mgr = cache1.getTransactionManager();
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      mgr.begin();
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.put(fqn, key, "value1");
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, key, "value2");
      mgr.commit();
      verifyReplication();
      // cache1 should still have this
      assertEquals("value2", cache1.get(fqn, key));

      if (!isInvalidation)
      {
         assertEquals("value1", cache2.get(fqn, key));
      }
      else
      {
         assertNull(cache2.get(fqn, key));
      }

      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      // now try again with passing the default options
      mgr.begin();
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put(fqn, key, "value3");
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.put(fqn, key, "value");
      mgr.commit();
      verifyReplication();
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));

      // cache 2 should as well
      if (!isInvalidation)
      {
         assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      mgr = cache2.getTransactionManager();
      mgr.begin();
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      cache2.put(fqn, key, "value3");
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, key, "value2");
      mgr.commit();
      verifyReplication();

      assertEquals("value2", cache2.get(fqn, key));

      if (!isInvalidation)
      {
         assertEquals("value3", cache1.get(fqn, key));
      }
      else
      {
         assertNull(cache1.get(fqn, key));
      }

      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      mgr.begin();
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache2.put(fqn, key, "value2");
      cache2.getInvocationContext().getOptionOverrides().reset();
      cache2.put(fqn, key, "value4");
      mgr.commit();
      verifyReplication();
      assertEquals("value4", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         assertEquals("value4", cache1.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache1.get(fqn, key));
      }

   }

   public void testTransactionalBehaviourRollback() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();

      TransactionManager mgr = cache1.getTransactionManager();

      // create these first ...
      cache1.put("/a", key, "old");
      cache1.put("/b", key, "old");
      Thread.sleep(500);


      mgr.begin();
      cache1.getInvocationContext().getOptionOverrides().reset();
      cache1.put("/a", key, "value1");
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      cache1.put("/b", key, "value2");
      mgr.rollback();
      Thread.sleep(500);
      // cache1 should NOT have this
      assert cache1.get("/a", key).equals("old");
      assert cache1.get("/b", key).equals("old");

      if (isInvalidation)
      {
         assert cache2.get("/a", key) == null;
         assert cache2.get("/b", key) == null;
      }
      else
      {
         assert cache2.get("/a", key).equals("old");
         assert cache2.get("/b", key).equals("old");
      }
   }

   public void testTransactionalBehaviourViaNodeAPI() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();
      NodeSPI<String, String> root1 = root1TL.get();
      NodeSPI<String, String> root2 = root2TL.get();

      Node node1 = root1.addChild(fqn);
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      TransactionManager mgr = cache1.getTransactionManager();
      mgr.begin();
      cache1.getInvocationContext().getOptionOverrides().reset();
      node1.put(key, "value1");
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node1.put(key, "value2");
      mgr.commit();
      verifyReplication();
      // cache1 should still have this
      assertEquals("value2", cache1.get(fqn, key));

      if (!isInvalidation)
      {
         assertEquals("value1", cache2.get(fqn, key));
      }
      else
      {
         assertNull(cache2.get(fqn, key));
      }

      // now try again with passing the default options
      registerReplicationCommand(cache2, PutKeyValueCommand.class);
      mgr.begin();
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node1.put(key, "value3");
      cache1.getInvocationContext().getOptionOverrides().reset();
      node1.put(key, "value");
      mgr.commit();
      verifyReplication();
      // cache1 should still have this
      assertEquals("value", cache1.get(fqn, key));

      // cache 2 should as well
      if (!isInvalidation)
      {
         assertEquals("value", cache2.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache2.get(fqn, key));
      }

      // now cache2
      Node node2 = root2.addChild(fqn);
      mgr = cache2.getTransactionManager();
      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      mgr.begin();
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);
      node2.put(key, "value3");
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node2.put(key, "value2");
      mgr.commit();
      verifyReplication();

      assertEquals("value2", cache2.get(fqn, key));

      if (!isInvalidation)
      {
         assertEquals("value3", cache1.get(fqn, key));
      }
      else
      {
         assertNull(cache1.get(fqn, key));
      }

      registerReplicationCommand(cache1, PutKeyValueCommand.class);
      mgr.begin();
      cache2.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      node2.put(key, "value2");
      cache2.getInvocationContext().getOptionOverrides().reset();
      node2.put(key, "value4");
      mgr.commit();
      verifyReplication();
      assertEquals("value4", cache2.get(fqn, key));
      if (!isInvalidation)
      {
         assertEquals("value4", cache1.get(fqn, key));
      }
      else
      {
         assertNull("should be invalidated", cache1.get(fqn, key));
      }

   }

   public void testAddChild() throws Exception
   {
      CacheSPI<String, String> cache1 = cache1TL.get();
      CacheSPI<String, String> cache2 = cache2TL.get();
      NodeSPI<String, String> root1 = root1TL.get();
      NodeSPI<String, String> root2 = root2TL.get();

      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(true);
      root1.addChild(fqn);
      Thread.sleep(1000);
      // cache1 should still have this
      assertTrue(root1.hasChild(fqn));
      // cache 2 should not
      Node node2 = root2.getChild(fqn);
      assertTrue("Should be null", node2 == null || (isInvalidation && !node2.isValid()));

      // now try again with passing the default options
      root1.removeChild(fqn);
      cache1.getInvocationContext().getOptionOverrides().setCacheModeLocal(false);

      registerReplicationCommand(cache2, PutDataMapCommand.class);
      root1.addChild(fqn);
      verifyReplication();

      // cache1 should still have this
      assertTrue(root1.hasChild(fqn));
      // cache 2 should as well
      if (!isInvalidation)
      {
         assertTrue(root2.hasChild(fqn));
      }
      else
      {
         assertTrue("Should be null", node2 == null || !node2.isValid());
      }
   }

   /**
    * Register a command that should be received by the specified cache as the result of a replication taking place.
    * @param cache the cache on which the given command should be replicated.
    * @param what the command that should be replicated
    * @see org.jboss.cache.options.cachemodelocal.AsyncReplPessLocksTest#registerReplicationCommand(org.jboss.cache.CacheSPI, Class)
    */
   protected abstract void registerReplicationCommand(CacheSPI<String, String> cache, Class<? extends ReplicableCommand> what);


   /**
    * Wait a while until the give command gets replicated.
    * @see AsyncReplPessLocksTest#verifyReplication() 
    */
   protected abstract void verifyReplication();
}